/*
  Copyright (C) 2009 Julien Jorge

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

  contact: julien.jorge@gamned.org
*/
#ifndef __SCRIPT_BASE_EXPORTABLE_HPP__
#define __SCRIPT_BASE_EXPORTABLE_HPP__

/** On va dfinir ci-dessous la classe de base pour les objets pouvant recevoir
    des commandes. Ces classes contiennent une table statique des
    method_caller  appeler pour un nom de mthode donne. Ce sera dfinit d'une
    manire qui sera, au final, invisible pour le crateur des classes
    drives. Ce dernier n'aura que quelques macros  appeler  la manire des
    vnements de wxWidgets.
*/

#include "script/method_caller.hpp"
#include "script/method_caller_args.hpp"

#include <string>
#include <iostream>
#include <map>

namespace script
{
  /*--------------------------------------------------------------------------*/
  /** La liste des mthode appelables. */
  template<typename Context>
  struct method_list
  {
    typedef std::map< std::string, method_caller<Context> const* > list_type;

    method_list() : parent(NULL) {}

    /** La liste de la classe mre. Cela implique qu'il n'y a pas d'hritage
        multiple pour l'instant. */
    method_list<Context> const* parent;

    /** La table des excuteurs. */
    list_type data;

  }; // struct method_list

  /*--------------------------------------------------------------------------*/
  /** Cette macro, dclare les variables membres statiques utilises pour
      stocker les excuteurs. Elle est appele automatiquement par
      DECLARE_METHOD_LIST. */
#define DECLARE_METHOD_LIST_BASE                                        \
  protected:                                                            \
  typedef script::method_list<script_context_type> method_list_type;    \
  typedef                                                               \
  std::map< std::string, script::method_caller<script_context_type> const* > \
  method_list_data_type;                                                \
                                                                        \
  static method_list_type s_method_list;                                \
                                                                        \
  static void self_methods_set                                          \
  ( const std::string& name,                                            \
    script::method_caller<script_context_type> const* m )               \
  { s_method_list.data[name] = m; }                                     \
                                                                        \
  private:                                                              \
  virtual method_list_type const* get_method_list() const               \
  {                                                                     \
    init_method_list();                                                 \
    return &s_method_list;                                              \
  }

  /*--------------------------------------------------------------------------*/
  /** Cette macro, appele dans la dfinition d'une classe, dclare les
      variables membres statiques utilises pour stoker les excuteurs.

      \param parent_class Le type de la classe mre (hritant de ou tant
             base_exportable)
      \param export_func Une mthode statique  appeler pour initialiser
             la liste des excuteurs.
  */
#define DECLARE_METHOD_LIST(parent_class, export_func)                  \
  DECLARE_METHOD_LIST_BASE                                              \
                                                                        \
  static void init_method_list()                                        \
  {                                                                     \
    if (s_method_list.parent == NULL)                                   \
      {                                                                 \
        parent_class::init_method_list();                               \
        s_method_list.parent = &parent_class::s_method_list;            \
        export_func();                                                  \
      }                                                                 \
  }

 /*--------------------------------------------------------------------------*/
  /** Cette macro est  utiliser dans le fichier d'implmentation de votre
      classe. */
 #define IMPLEMENT_METHOD_LIST(self_type)        \
  script::method_list<self_type::script_context_type> self_type::s_method_list;

  /*--------------------------------------------------------------------------*/
  /** Cette macro est utilise dans la mthode statique passe 
      DECLARE_METHOD_LIST pour ajouter un excuteur d'une mthode sans
      paramtres.
  */
#define CONNECT_METHOD_0( self_type, method_name, R )           \
  self_methods_set                                              \
  ( #method_name,                                               \
    &script::method_caller_args_0                               \
    <                                                           \
    self_type,                                                  \
    R,                                                          \
    &self_type::method_name>::s_caller )

  /*--------------------------------------------------------------------------*/
  /** Cette macro est utilise dans la mthode statique passe 
      DECLARE_METHOD_LIST pour ajouter un excuteur d'une mthode ayant un
      unique paramtre, de type T1. */
#define CONNECT_METHOD_1( self_type, method_name, R, T1 )	\
  self_methods_set                                              \
  ( #method_name,                                               \
    &script::method_caller_args_1                               \
    <                                                           \
    self_type,                                                  \
    R, T1,                                                      \
    &self_type::method_name >::s_caller )

  /*--------------------------------------------------------------------------*/
  /** Cette macro est utilise dans la mthode statique passe 
      DECLARE_METHOD_LIST pour ajouter un excuteur d'une mthode ayant deux
      paramtres, de type T1 et T2. */
#define CONNECT_METHOD_2( self_type, method_name, R, T1, T2 )	\
  self_methods_set                                              \
  ( #method_name,                                               \
    &script::method_caller_args_2                               \
    <                                                           \
    self_type,                                                  \
    R, T1, T2,                                                  \
    &self_type::method_name >::s_caller )

  /*--------------------------------------------------------------------------*/
  /** Cette macro est utilise dans la mthode statique passe 
      DECLARE_METHOD_LIST pour ajouter un excuteur d'une mthode ayant trois
      paramtres, de type T1, T2 et T3. */
#define CONNECT_METHOD_3( self_type, method_name, R, T1, T2, T3 )	\
  self_methods_set                                                      \
  ( #method_name,                                                       \
    &script::method_caller_args_3					\
    <									\
    self_type,                                                          \
    R, T1, T2, T3,							\
    &self_type::method_name >::s_caller )

  /*--------------------------------------------------------------------------*/
  /** Cette classe est la base de toutes les classes pouvant excuter des
      commandes. */
  template<typename Context>
  class base_exportable
  {
  public:
    // Le type du contexte. Dclaration indispensable pour l'utilisation de
    // DECLARE_METHOD_LIST_BASE ci-aprs.
    typedef Context script_context_type;

    // Nous utilisons ici, et uniquement ici, la version _BASE car il n'y a pas
    // de classe mre dans laquelle d'autres mthodes seraient exportes.
    DECLARE_METHOD_LIST_BASE

  protected:
    // Il n'y a rien  exporter de cette classe.
    static void init_method_list() {}

  public:
    /** Destructeur indispensable pour pouvoir utiliser le dynamic_cast dans
        explicit_method_caller. */
    virtual ~base_exportable() {}

    /** Excute une mthode de la classe. \a n est le nom de la mthode, \a args
        ses paramtres. \a context est le contexte d'excution. */
    void execute( const std::string& n, const std::vector<std::string>& args,
                  const script_context_type& context )
    {
      // on rcupre la liste des excuteurs dans la classe la plus basse dans
      // la hirarchie.
      const method_list_type* m( get_method_list() );
      typename method_list_data_type::const_iterator it(m->data.find(n));

      bool stop(false);

      while( !stop )
        if ( it == m->data.end() )
          {
            // On n'a pas trouv la mthode, on passe alors  la
            // liste de la classe mre, si elle existe.
            if ( m->parent != NULL )
              {
                m = m->parent;
                it = m->data.find(n);
              }
            else
              {
                std::cout << "Method '" << n << "' not found." << std::endl;
                stop = true;
              }
          }
        else
          {
            // On a trouv la bonne mthode, on l'excute.
            stop = true;
            it->second->execute(this, args, context);
          }
    }
  
  }; // class base_exportable

  template<typename Context>
  method_list<Context> base_exportable<Context>::s_method_list;

} // namespace script

#endif // __SCRIPT_BASE_EXPORTABLE_HPP__
